#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Secrets Manager
# Copyright 2021 Keeper Security Inc.
# Contact: ops@keepersecurity.com

import hmac
import logging
import os
from distutils.util import strtobool
import re
import json

import requests
from requests import HTTPError

from keeper_secrets_manager_core import utils, helpers
from keeper_secrets_manager_core.configkeys import ConfigKeys
from keeper_secrets_manager_core.dto.dtos import Folder, Record
from keeper_secrets_manager_core.dto.payload import GetPayload, UpdatePayload, Context, TransmissionKey
from keeper_secrets_manager_core.exceptions import KeeperError
from keeper_secrets_manager_core.keeper_globals import keeper_secrets_manager_sdk_client_id, keeper_public_keys, \
    logger_name
from keeper_secrets_manager_core.storage import FileKeyValueStorage, KeyValueStorage
from keeper_secrets_manager_core.utils import bytes_to_url_safe_str, base64_to_bytes, sign, \
    extract_public_key_bytes, dict_to_json, url_safe_str_to_bytes, encrypt_aes, der_base64_private_key_to_private_key, \
    string_to_bytes, decrypt_aes, bytes_to_string, json_to_dict, \
    public_encrypt, generate_private_key_der


class SecretsManager:

    notation_prefix = "keeper"
    default_key_id = "7"

    def __init__(self, token=None, hostname=None, verify_ssl_certs=True, config=None, log_level=None):

        self.token = token
        self.hostname = hostname

        # Init the log, create a logger for the core.
        self._init_logger(log_level=log_level)
        self.logger = logging.getLogger(logger_name)

        # Accept the env var KSM_SKIP_VERIFY. Modules like 'requests' already use it.
        self.verify_ssl_certs = verify_ssl_certs
        if os.environ.get("KSM_SKIP_VERIFY") is not None:
            # We need to flip the value of KSM_SKIP_VERIFY, if true, we want verify_ssl_certs to be false.
            self.verify_ssl_certs = not bool(strtobool(os.environ.get("KSM_SKIP_VERIFY")))

        if config is None:
            config = FileKeyValueStorage()

        # If the server or client key are set in the args, make sure they makes it's way into the config. They
        # will override what is already in the config if they exist.
        if token is not None:
            config.set(ConfigKeys.KEY_CLIENT_KEY, token)
        if hostname is not None:
            config.set(ConfigKeys.KEY_HOSTNAME, hostname)

        # Make sure our public key id is set and pointing an existing key.
        if config.get(ConfigKeys.KEY_SERVER_PUBLIC_KEY_ID) is None:
            self.logger.debug("Setting public key id to the default: {}".format(SecretsManager.default_key_id))
            config.set(ConfigKeys.KEY_SERVER_PUBLIC_KEY_ID, SecretsManager.default_key_id)
        elif config.get(ConfigKeys.KEY_SERVER_PUBLIC_KEY_ID) not in keeper_public_keys:
            self.logger.debug("Public key id {} does not exists, set to default : {}".format(
                config.get(ConfigKeys.KEY_SERVER_PUBLIC_KEY_ID),
                SecretsManager.default_key_id))
            config.set(ConfigKeys.KEY_SERVER_PUBLIC_KEY_ID, SecretsManager.default_key_id)

        self.config: KeyValueStorage = config

        self._init()

    @staticmethod
    def _init_logger(log_level=None):

        logger = logging.getLogger(logger_name)

        # If the log level was passed in, then we want to set up the logging. If not, there is no logging.
        if log_level is not None:

            # If the log level is a string, get the enum value.
            if type(log_level) is str:
                valid_log_levels = ["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG", "NOTSET"]
                if log_level not in valid_log_levels:
                    raise ValueError("Log level {} is invalid. Valid values are: {}".format(
                        log_level, ", ".join(valid_log_levels)))
                log_level = getattr(logging, log_level)

            logger.disabled = False
            logger.setLevel(log_level)

            # If we don't have my handlers, set the logger up with default settings. The handler and formatters
            # can be set outside of the SDK if an integration or company
            if len(logger.handlers) == 0:
                # Default to stderr
                sh = logging.StreamHandler()
                sh.setLevel(log_level)

                sh.setFormatter(logging.Formatter("%(asctime)s | %(name)s | %(levelname)s | %(message)s"))
                logger.addHandler(sh)
        else:
            logger.disabled = True

    def _init(self):

        client_id = self.config.get(ConfigKeys.KEY_CLIENT_ID)

        if client_id:
            self.logger.debug("Already bound")

            if self.config.get(ConfigKeys.KEY_CLIENT_KEY):
                self.config.delete(ConfigKeys.KEY_CLIENT_KEY)

        else:

            existing_secret_key = self.load_secret_key()

            if existing_secret_key is None:
                raise ValueError("Cannot locate One Time Token.")

            existing_secret_key_bytes = url_safe_str_to_bytes(existing_secret_key)
            digest = 'sha512'
            existing_secret_key_hash = bytes_to_url_safe_str(hmac.new(existing_secret_key_bytes,
                                                                      b'KEEPER_SECRETS_MANAGER_CLIENT_ID',
                                                                      digest).digest())

            self.config.delete(ConfigKeys.KEY_CLIENT_ID)
            self.config.delete(ConfigKeys.KEY_PRIVATE_KEY)
            self.config.delete(ConfigKeys.KEY_APP_KEY)

            self.config.set(ConfigKeys.KEY_CLIENT_ID, existing_secret_key_hash)

            private_key_str = self.config.get(ConfigKeys.KEY_PRIVATE_KEY)

            if not private_key_str:
                private_key_der = generate_private_key_der()
                self.config.set(ConfigKeys.KEY_PRIVATE_KEY, bytes_to_url_safe_str(private_key_der))

        if not self.verify_ssl_certs:
            self.logger.warning("WARNING: Running without SSL cert verification. "
                                "Execute 'SecretsManager(..., verify_ssl_certs=True)' or 'KSM_SKIP_VERIFY=FALSE' "
                                "to enable verification.")

    def load_secret_key(self):

        """Returns client_id from the environment variable, config file, or in the code"""

        # Case 1: Environment Variable
        env_secret_key = os.getenv('KSM_TOKEN')

        current_secret_key = None

        if env_secret_key:
            current_secret_key = env_secret_key
            self.logger.info("Secret key found in environment variable")

        # Case 2: Code
        if not current_secret_key:
            code_secret_key = self.token

            if code_secret_key:
                current_secret_key = code_secret_key
                self.logger.info("Secret key found in code")

        # Case 3: Config storage
        if not current_secret_key:
            config_secret_key = self.config.get(ConfigKeys.KEY_CLIENT_KEY)

            if config_secret_key:
                current_secret_key = config_secret_key
                self.logger.info("Secret key found in configuration file")

        return current_secret_key

    @staticmethod
    def generate_transmission_key(key_id):
        transmission_key = utils.generate_random_bytes(32)

        if key_id not in keeper_public_keys:
            ValueError("The public key id {} does not exist.".format(key_id))

        server_public_raw_key_bytes = url_safe_str_to_bytes(keeper_public_keys[key_id])

        encrypted_key = public_encrypt(transmission_key, server_public_raw_key_bytes)

        return TransmissionKey(key_id, transmission_key, encrypted_key)

    def prepare_context(self):

        # Get the index of the public get we need to use. If not set, set it the default and
        # save it back into the config
        key_id = self.config.get(ConfigKeys.KEY_SERVER_PUBLIC_KEY_ID)
        if key_id is None:
            key_id = SecretsManager.default_key_id
            self.config.set(ConfigKeys.KEY_SERVER_PUBLIC_KEY_ID, str(key_id))

        # Generate the transmission key using the public key at the key id index
        transmission_key = SecretsManager.generate_transmission_key(key_id)
        client_id = self.config.get(ConfigKeys.KEY_CLIENT_ID)
        secret_key = None

        # While not use in the normal operations, it's used for mocking unit tests.
        app_key = self.config.get(ConfigKeys.KEY_APP_KEY)
        if app_key is not None:
            secret_key = base64_to_bytes(app_key)

        if not client_id:
            raise Exception("Client ID is missing from the configuration")

        client_id_bytes = base64_to_bytes(client_id)

        return Context(
                    transmission_key,
                    client_id_bytes,
                    secret_key
                )

    def encrypt_and_sign_payload(self, context, payload):

        if not (isinstance(payload, GetPayload) or isinstance(payload, UpdatePayload)):
            raise Exception('Unknown payload type "%s"' % payload.__class__.__name__)

        payload_json_str = dict_to_json(payload.__dict__)
        payload_bytes = string_to_bytes(payload_json_str)

        encrypted_payload = encrypt_aes(payload_bytes, context.transmissionKey.key)

        encrypted_key = context.transmissionKey.encryptedKey
        signature_base = encrypted_key + encrypted_payload

        pk = der_base64_private_key_to_private_key(self.config.get(ConfigKeys.KEY_PRIVATE_KEY))
        signature = sign(signature_base, pk)

        return {
            'payload':  encrypted_payload,
            'signature': signature
        }

    def prepare_get_payload(self, context, records_filter):

        payload = GetPayload()

        payload.clientVersion = keeper_secrets_manager_sdk_client_id
        payload.clientId = bytes_to_url_safe_str(context.clientId) + "="

        app_key_str = self.config.get(ConfigKeys.KEY_APP_KEY)

        if not app_key_str:

            public_key_bytes = extract_public_key_bytes(self.config.get(ConfigKeys.KEY_PRIVATE_KEY))
            public_key_base64 = bytes_to_url_safe_str(public_key_bytes)
            # passed once when binding
            payload.publicKey = public_key_base64 + "="

        if records_filter:
            payload.requestedRecords = records_filter

        return self.encrypt_and_sign_payload(context, payload)

    def prepare_update_payload(self, context, record):
        payload = UpdatePayload()

        payload.clientVersion = keeper_secrets_manager_sdk_client_id
        payload.clientId = bytes_to_url_safe_str(context.clientId) + "="

        if not context.clientKey:
            raise KeeperError("To save and update, client must be authenticated by device token only")

        # for update, uid of the record
        payload.recordUid = record.uid

        # TODO: This is where we need to get JSON of the updated Record
        raw_json_bytes = string_to_bytes(record.raw_json)
        encrypted_raw_json_bytes = encrypt_aes(raw_json_bytes, record.record_key_bytes)
        encrypted_raw_json_bytes_str = bytes_to_url_safe_str(encrypted_raw_json_bytes)

        # for create and update, the record data
        payload.data = encrypted_raw_json_bytes_str

        return self.encrypt_and_sign_payload(context, payload)

    def _post_query(self, path, context, payload_and_signature):

        keeper_server = helpers.get_server(self.hostname, self.config)

        transmission_key = context.transmissionKey
        payload = payload_and_signature.get('payload')
        signature = payload_and_signature.get('signature')

        request_headers = {
            'Content-Type': 'application/octet-stream', 'Content-Length': str(len(payload)),
            'PublicKeyId': str(transmission_key.publicKeyId),
            'TransmissionKey': bytes_to_url_safe_str(transmission_key.encryptedKey),
            'Authorization': 'Signature %s' % bytes_to_url_safe_str(signature)
        }

        rs = requests.post(
            'https://%s/api/rest/sm/v1/%s' % (keeper_server, path),
            headers=request_headers,
            data=payload,
            verify=self.verify_ssl_certs
        )

        return rs

    def handler_http_error(self, rs):

        log_level = logging.ERROR
        try:
            # Decode the JSON content, throw exception if not JSON
            response_dict = json_to_dict(rs.text)
            if response_dict is None:
                raise json.JSONDecodeError("Not JSON", "NONE", 0)

            # Try to get the error from result_code, then from error.
            error = response_dict.get('result_code', response_dict.get('error'))

            if error == 'invalid_client_version':
                self.logger.error("Client version {} was not registered in the backend".format(
                    keeper_secrets_manager_sdk_client_id))
                msg = response_dict.get('additional_info')

            # The server wants us to use a different public key.
            elif error == 'key':
                key_id = response_dict.get("key_id")
                self.logger.info("Server has requested we use public key {}".format(key_id))

                if key_id is None:
                    raise ValueError("The public key is blank from the server")
                elif str(key_id) not in keeper_public_keys:
                    raise ValueError("The public key at {} does not exist in the SDK".format(key_id))

                self.config.set(ConfigKeys.KEY_SERVER_PUBLIC_KEY_ID, str(key_id))

                # This is not an error, it's info. Make it so the 'finally' display info about the
                # key change.
                log_level = logging.DEBUG

                # The only non-exception exit from this method
                return True
            else:
                msg = "Error: {}, message={}".format(error, response_dict.get('message', "NA"))

            raise KeeperError(msg)

        except json.JSONDecodeError as _:
            # The content wasn't JSON. Let the catch-all exception at the end handle it.
            pass
        except KeeperError as err:
            # This was one of our exceptions, just rethrow it.
            raise err
        finally:
            self.logger.log(log_level, "Error: {} (http error code {}): {}".format(
                rs.reason,
                rs.status_code,
                rs.text
            ))

        # This is a unknown error, not one of ours, just throw a HTTPError
        raise HTTPError(rs.text)

    def fetch(self, record_filter=None):

        while True:

            context = self.prepare_context()
            payload_and_signature = self.prepare_get_payload(context, records_filter=record_filter)

            rs = self._post_query(
                'get_secret',
                context,
                payload_and_signature
            )

            # If we are ok, then break out of the while loop
            if rs.ok:
                break

            # Handle the error. Handling will throw an exception if it doesn't want us to retry.
            self.handler_http_error(rs)

        decrypted_response_bytes = decrypt_aes(rs.content, context.transmissionKey.key)
        decrypted_response_str = bytes_to_string(decrypted_response_bytes)
        decrypted_response_dict = json_to_dict(decrypted_response_str)

        records = []

        just_bound = False

        if decrypted_response_dict.get('encryptedAppKey'):
            just_bound = True

            encrypted_master_key = url_safe_str_to_bytes(decrypted_response_dict.get('encryptedAppKey'))
            secret_key = decrypt_aes(encrypted_master_key, self.config.get(ConfigKeys.KEY_CLIENT_KEY))
            self.config.set(ConfigKeys.KEY_APP_KEY, bytes_to_url_safe_str(secret_key))

            self.config.delete(ConfigKeys.KEY_CLIENT_KEY)

        else:
            secret_key = base64_to_bytes(self.config.get(ConfigKeys.KEY_APP_KEY))

        records_resp = decrypted_response_dict.get('records')
        folders_resp = decrypted_response_dict.get('folders')

        self.logger.debug("Individual record count: {}".format(len(records_resp or [])))
        self.logger.debug("Folder count: {}".format(len(folders_resp or [])))

        if records_resp:
            for r in records_resp:
                record = Record(r, secret_key)
                records.append(record)

        if folders_resp:
            for f in folders_resp:
                folder = Folder(f, secret_key)
                records.extend(folder.records)

        self.logger.debug("Total record count: {}".format(len(records)))

        return {
            'records': records,
            'justBound': just_bound
        }

    def get_secrets(self, uids=None):
        """
        Retrieve all records associated with the given application
        """
        records_resp = self.fetch(uids)

        just_bound = records_resp.get('justBound')

        if just_bound:
            records_resp = self.fetch(uids)

        # TODO: Erase client key because we are already bound

        records = records_resp.get('records') or []

        return records

    def save(self, record):
        """
        Save updated secret values
        """

        self.logger.info("Updating record uid: %s" % record.uid)

        while True:

            context = self.prepare_context()
            payload_and_signature = self.prepare_update_payload(context, record)

            rs = self._post_query(
                'update_secret',
                context,
                payload_and_signature
            )

            # If we are ok, then break out of the while loop
            if rs.ok:
                break

            # Handle the error. Handling will throw an exception if it doesn't want us to retry.
            self.handler_http_error(rs)

        return True

    def get_notation(self, url):

        """Simple string notation to get a value

        * A system of figures or symbols used in a specialized field to represent numbers, quantities, tones,
          or values.

        <uid>/<field|custom_field|file>/<label|type>[INDEX][FIELD]

        Example:

            EG6KdJaaLG7esRZbMnfbFA/field/password                => MyPasswprd
            EG6KdJaaLG7esRZbMnfbFA/field/password[0]             => MyPassword
            EG6KdJaaLG7esRZbMnfbFA/field/password[]              => ["MyPassword"]
            EG6KdJaaLG7esRZbMnfbFA/custom_field/name[first]      => John
            EG6KdJaaLG7esRZbMnfbFA/custom_field/name[last]       => Smitht
            EG6KdJaaLG7esRZbMnfbFA/custom_field/phone[0][number] => "555-5555555"
            EG6KdJaaLG7esRZbMnfbFA/custom_field/phone[1][number] => "777-7777777"
            EG6KdJaaLG7esRZbMnfbFA/custom_field/phone[]          => [{"number": "555-555...}, { "number": "777.....}]
            EG6KdJaaLG7esRZbMnfbFA/custom_field/phone[0]         => [{"number": "555-555...}]

        """

        # If the URL starts with keeper:// we want to remove it.
        if url.startswith(SecretsManager.notation_prefix) is True:
            url_parts = url.split('//')
            try:
                url = url_parts[1]
                if url is None:
                    # Get the except below handle it
                    raise ValueError()
            except IndexError:
                raise ValueError("Keeper url missing information about the uid, field type, and field key.")

        try:
            (uid, file_type, key) = url.split('/')
        except Exception as _:
            raise ValueError("Could not parse the notation {}. Is it valid?".format(url))

        if uid is None:
            raise ValueError("UID is missing the in the keeper url.")
        if file_type is None:
            raise ValueError("file type is missing the in the keeper url.")
        if key is None:
            raise ValueError("file key is missing the in the keeper url.")

        # By default we want to return a single value, which is the first item in the array
        return_single = True
        index = 0
        dict_key = None

        # Check it see if the key has a predicate, possibly with an index.
        predicate = re.search(r'\[.*]', key)
        if predicate is not None:

            # If we do, get the predicate and remove the brackets, to get the index if one exists
            match = predicate.group()

            predicate_parts = match.split("]")
            while "" in predicate_parts:
                predicate_parts.remove("")

            if len(predicate_parts) == 0:
                raise ValueError("The predicate of the notation appears to be invalid. Syntax error?")
            if len(predicate_parts) > 2:
                raise ValueError("The predicate of the notation appears to be invalid. Too many [], max 2 allowed.")

            # This will remove the preceding '['
            first_predicate = predicate_parts[0][1:]
            if first_predicate is not None:
                # Is the first predicate an index into an array?
                if first_predicate.isdigit() is True:
                    index = int(first_predicate)
                # Is the first predicate a key to a dictionary?
                elif re.match(r'^[a-zA-Z0-9_]+$', first_predicate):
                    dict_key = first_predicate
                # Else it was an array indicator. Return all the values.
                else:
                    return_single = False

            if len(predicate_parts) == 2:
                if return_single is False:
                    raise ValueError("If the second [] is a dictionary key, the first [] needs to have any index.")
                # Remove the preceding '['
                second_predicate = predicate_parts[1][1:]
                if second_predicate.isdigit() is True:
                    raise ValueError("The second [] can only by a key for the dictionary. It cannot be an index.")
                # Is the first predicate a key to a dictionary?
                elif re.match(r'^[a-zA-Z0-9_]+$', second_predicate):
                    dict_key = second_predicate
                else:
                    raise ValueError("The second [] must have key for the dictionary. Cannot be blank.")

            # Remove the predicate from the key, if it exists
            key = re.sub(r'\[.*', '', key)

        records = self.get_secrets([uid])
        if len(records) == 0:
            raise ValueError("Could not find a record with the UID {}".format(uid))

        record = records[0]

        if file_type == "field":
            value = record.field(key, single=False)
        elif file_type == "custom_field":
            value = record.custom_field(key, single=False)
        elif file_type == "file":
            file = record.find_file_by_title(key)
            if file is None:
                raise FileNotFoundError("Cannot find the file {} in record {}.".format(key, uid))
            value = file.get_file_data()
        else:
            raise ValueError("Field type of {} is not value.".format(file_type))

        ret = value
        if return_single is True and type(value) is list:
            if len(value) == 0:
                return None
            try:
                ret = value[index]
                if dict_key is not None:
                    if dict_key not in ret:
                        raise ValueError("Cannot find the dictionary key {} in the valuye".format(dict_key))
                    ret = ret[dict_key]
            except IndexError:
                raise ValueError("The value at index {} does not exist for {}.".format(index, url))
        return ret



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Secrets Manager
# Copyright 2021 Keeper Security Inc.
# Contact: ops@keepersecurity.com

import base64
import json
import logging
import os
from json import JSONDecodeError
from sys import platform as _platform

from Cryptodome.Cipher import AES, PKCS1_v1_5
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization, hashes
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.serialization import load_der_private_key
from keeper_secrets_manager_core.keeper_globals import logger_name

ENCODING = 'UTF-8'


def get_os():
    if _platform.lower().startswith("linux"):
        return "linux"
    elif _platform.lower().startswith("darwin"):
        return "macOS"
    # elif _platform.lower().startswith("win32"):
    #     return "win32"
    # elif _platform.lower().startswith("win64"):
    #     return "win64"
    else:
        return _platform


def bytes_to_string(b):
    return b.decode(ENCODING)


def bytes_to_int(b):
    return int.from_bytes(b, byteorder='big')


def bytes_to_url_safe_str(b):
    return base64.urlsafe_b64encode(b).decode().rstrip('=')


def base64_to_bytes(s):
    bbytes = base64.urlsafe_b64decode(s + '==')

    return bbytes


def string_to_bytes(s):
    return s.encode(ENCODING)


def url_safe_str_to_bytes(s):
    b = base64.urlsafe_b64decode(s + '==')
    return b


def url_safe_str_to_int(s):
    b = url_safe_str_to_bytes(s)
    return bytes_to_int(b)


def generate_random_bytes(length):
    return os.urandom(length)


def dict_to_json(dictionary):
    return json.dumps(dictionary, indent=4)


def json_to_dict(json_str):

    try:
        resp = json.loads(json_str)
    except JSONDecodeError as jsonDecErr:
        logging.getLogger(logger_name).warning(jsonDecErr)
        resp = None

    return resp


BS = 16

pad_binary = lambda s: s + ((BS - len(s) % BS) * chr(BS - len(s) % BS)).encode()

unpad_binary = lambda s: s[0:-s[-1]]

unpad_char = lambda s: s[0:-ord(s[-1])]


########################################################################################################################
#                                                  Encryption methods                                                  #
########################################################################################################################


def generate_private_key_ecc():
    encryption_key_bytes = generate_random_bytes(32)
    private_key_str = bytes_to_url_safe_str(encryption_key_bytes)

    encryption_key_int = url_safe_str_to_int(private_key_str)

    private_key = ec.derive_private_key(encryption_key_int, ec.SECP256R1(), default_backend())

    return private_key


def generate_private_key_der():

    private_key = generate_private_key_ecc()

    # export to DER

    private_key_der = private_key.private_bytes(
        encoding=serialization.Encoding.DER,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption(),
    )

    return private_key_der


def generate_new_ecc_key():
    curve = ec.SECP256R1()
    ephemeral_key = ec.generate_private_key(curve, default_backend())
    return ephemeral_key


def encrypt_rsa(data, rsa_key):
    # type: (bytes, RSA.RsaKey) -> bytes
    cipher = PKCS1_v1_5.new(rsa_key)
    return cipher.encrypt(data)


def encrypt_aes(data, key):
    # type: (bytes, bytes) -> bytes
    iv = os.urandom(12)
    cipher = AES.new(key=key, mode=AES.MODE_GCM, nonce=iv)
    enc_data, tag = cipher.encrypt_and_digest(data)
    return iv + enc_data + tag


def public_encrypt(data: bytes, server_public_raw_key_bytes: bytes, idz: bytes = None):
    try:
        curve = ec.SECP256R1()

        ephemeral_public_key = ec.EllipticCurvePublicKey.from_encoded_point(curve, server_public_raw_key_bytes)

        # server_public_key = load_pem_public_key(server_public_key_bytes)
        ephemeral_key2 = generate_new_ecc_key()
        shared_key = ephemeral_key2.exchange(ec.ECDH(), ephemeral_public_key)

        if idz:
            shared_key = shared_key + idz

        digest = hashes.Hash(hashes.SHA256(), backend=default_backend())
        digest.update(shared_key)
        enc_key = digest.finalize()
        encrypted_data = encrypt_aes(data, enc_key)
        eph_public_key = ephemeral_key2.public_key().public_bytes(serialization.Encoding.X962,
                                                                  serialization.PublicFormat.UncompressedPoint)

        return eph_public_key + encrypted_data

    except Exception as e:
        logging.getLogger(logger_name).warning(e)
        return


def hash_of_string(value):
    value_bytes = base64_to_bytes(value)
    digest = hashes.Hash(hashes.SHA256(), backend=default_backend())
    digest.update(value_bytes)
    return digest.finalize()


def ecies_decrypt(server_public_key, ciphertext, priv_key_data, id=b''):

    try:
        # server_public_key = ciphertext[:65]
        encrypted_data = ciphertext
        curve = ec.SECP256R1()
        private_value = int.from_bytes(priv_key_data, byteorder='big', signed=False)
        ecc_private_key = ec.derive_private_key(private_value, curve, default_backend())

        # ecc_private_key = CommonHelperMethods.get_private_key_ecc(params)
        ephemeral_public_key = ec.EllipticCurvePublicKey.from_encoded_point(curve, server_public_key)
        shared_key = ecc_private_key.exchange(ec.ECDH(), ephemeral_public_key)
        shared_key = shared_key + id if id else shared_key
        digest = hashes.Hash(hashes.SHA256(), backend=default_backend())
        digest.update(shared_key)
        enc_key = digest.finalize()
        result = decrypt_aes(encrypted_data, enc_key)

    except Exception as e:
        raise e

    return result


def decrypt_data_aes(data, key):
    # type: (str, bytes) -> bytes
    decoded_data = base64.urlsafe_b64decode(data + '==')
    iv = decoded_data[:16]
    ciphertext = decoded_data[16:]
    cipher = AES.new(key, AES.MODE_GCM, iv)
    return cipher.decrypt(ciphertext)


def decrypt_aes(encrypted_data, key):
    # type: (bytes, bytes) -> bytes

    if isinstance(key, str):
        key = base64_to_bytes(key)

    nonce = encrypted_data[:12]
    cipher = AES.new(key=key, mode=AES.MODE_GCM, nonce=nonce)
    return cipher.decrypt_and_verify(encrypted_data[12:-16], encrypted_data[-16:])


def decrypt_record(data, secret_key):
    if isinstance(data, str):
        data = base64_to_bytes(data)

    record = decrypt_aes(data, secret_key)
    record_json = bytes_to_string(record)
    return record_json


def decrypt_data(data, key):
    # type: (str, bytes) -> bytes

    data_padded = decrypt_data_aes(data, key)
    un_padded = unpad_binary(data_padded)
    return un_padded


def decrypt_ec(ecc_private_key, encrypted_data_bag: bytes):
    curve = ec.SECP256R1()

    server_public_key = encrypted_data_bag[:65]
    encrypted_data = encrypted_data_bag[65:]

    ephemeral_public_key = ec.EllipticCurvePublicKey.from_encoded_point(curve, server_public_key)
    shared_key = ecc_private_key.exchange(ec.ECDH(), ephemeral_public_key)
    digest = hashes.Hash(hashes.SHA256(), backend=default_backend())
    digest.update(shared_key)
    enc_key = digest.finalize()
    decrypted_data = decrypt_aes(encrypted_data, enc_key)
    return decrypted_data


def der_base64_private_key_to_private_key(private_key_der_base64):

    if isinstance(private_key_der_base64, str):
        private_key_der_base64 = base64_to_bytes(private_key_der_base64)

    return load_der_private_key(private_key_der_base64, password=None)


def extract_public_key_bytes(private_key_der_base64):

    if isinstance(private_key_der_base64, str):
        private_key_der_base64 = base64_to_bytes(private_key_der_base64)

    ec_private_key = der_base64_private_key_to_private_key(private_key_der_base64)
    pub_key = ec_private_key.public_key()
    pub_key_bytes = pub_key.public_bytes(serialization.Encoding.X962, serialization.PublicFormat.UncompressedPoint)
    return pub_key_bytes


def sign(data, private_key):
    signature = private_key.sign(
        data,
        ec.ECDSA(hashes.SHA256())
    )
    return signature



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Secrets Manager
# Copyright 2021 Keeper Security Inc.
# Contact: ops@keepersecurity.com

import os

import requests

from keeper_secrets_manager_core.exceptions import KeeperError
from keeper_secrets_manager_core.utils import base64_to_bytes, decrypt_aes, decrypt_record, json_to_dict, dict_to_json


class Record:

    def __init__(self, record_dict, secret_key):
        self.uid = ''
        self.title = ''
        self.type = ''
        self.files = []
        self.raw_json = None
        self.dict = {}
        self.password = None

        self.uid = record_dict.get('recordUid')

        if 'recordKey' in record_dict and record_dict.get('recordKey'):
            # Folder Share
            record_key_encrypted_str = record_dict.get('recordKey')
            record_key_encrypted_bytes = base64_to_bytes(record_key_encrypted_str) if record_key_encrypted_str else None

            self.record_key_bytes = decrypt_aes(record_key_encrypted_bytes, secret_key)
        else:
            # Single Record Share
            self.record_key_bytes = secret_key

        record_encrypted_data = record_dict.get('data')
        record_data_json = decrypt_record(record_encrypted_data, self.record_key_bytes)

        self.raw_json = record_data_json
        self.dict = json_to_dict(self.raw_json)
        self.title = self.dict.get('title')
        self.type = self.dict.get('type')

        # files
        if record_dict.get('files'):
            for f in record_dict.get('files'):

                file = KeeperFile(f, self.record_key_bytes)

                self.files.append(file)

        # password (if `login` type)
        if self.type == 'login':

            fields = self.dict.get('fields')

            password_field = next((item for item in fields if item["type"] == "password"), None)

            # If the password field exists and there is a value in the array, then set the password.
            if password_field is not None and len(password_field.get('value', [])) > 0:
                self.password = password_field.get('value')[0]

    def find_file_by_title(self, title):
        """Finds file by file title"""

        found_file = next((f for f in self.files if f.title == title), None)

        return found_file

    def download_file_by_title(self, title, path, ):

        found_file = self.find_file_by_title(title)

        found_file.save_file(path)

    def __str__(self):
        return '[Record: uid=%s, type: %s, title: %s, files count: %s]' % (self.uid, self.type, self.title,
                                                                           str(len(self.files)))

    def _update(self):

        """ Take the values in the diction and update the attributes and raw JSON
        """

        self.dict["title"] = self.title
        self.dict["type"] = self.type

        # Find the password in the field and update the password attribute
        password_field = next((item for item in self.dict["fields"] if item["type"] == "password"), None)
        self.password = password_field.get('value')[0]

        self.raw_json = dict_to_json(self.dict)

    @staticmethod
    def _value(values, single):

        if single is True:
            return values[0]
        return values

    def field(self, field_type, value=None, single=False):

        found_item = None
        for item in self.dict.get('fields'):
            if item["type"] == field_type.lower():
                found_item = item
                break
        if found_item is None:
            raise ValueError("Cannot find the field '{}'.".format(field_type))

        if value is None:
            value = Record._value(found_item["value"], single)
        else:
            if type(value) is not list:
                value = [value]
            found_item["value"] = value
            self._update()

        return value

    def custom_field(self, label, value=None, field_type=None, single=False):

        found_item = None
        for item in self.dict.get('custom', []):
            found = False

            # If the user doesn't set the label in the UI, and uses the default, the label will be missing :/
            # Set the label to the type.
            if item.get("label") is None:
                item["label"] = item["type"]

            if item["label"] == label:
                # We can have duplicate labels, so allow type to be used too.
                if field_type is not None:
                    if item["type"] == field_type.lower():
                        found = True
                else:
                    found = True
            if found is True:
                found_item = item
                break

        if found_item is None:
            raise ValueError("Cannot find the custom field label='{}', field type='{}'.".format(label, field_type))

        if value is None:
            value = Record._value(found_item["value"], single)
        else:
            if type(value) is not list:
                value = [value]
            found_item["value"] = value
            self._update()

        return value

    def print(self):

        print("===")
        print("Title: {}".format(self.title))
        print("UID:   {}".format(self.uid))
        print("Type:  {}".format(self.type))
        print("")
        print("Fields")
        print("------")

        for item in self.dict.get('fields'):
            if item["type"] in ["fileRef", "oneTimeCode"]:
                continue
            print("{} : {}".format(item["type"], ", ".join(item["value"])))

        print("")
        print("Custom Fields")
        print("------")
        for item in self.dict.get('custom', []):
            print("{} ({}) : {}".format(item["label"], item["type"], ", ".join(item["value"])))


class Folder:

    def __init__(self, folder, secret_key):

        self.uid = ''
        self.records = []

        if not folder.get('folderUid'):
            raise Exception("Not a folder")

        folder_uid = folder.get('folderUid')
        folder_key_enc = folder.get('folderKey')
        folder_key = decrypt_aes(base64_to_bytes(folder_key_enc), secret_key)
        folder_records = folder.get('records')

        self.uid = folder_uid
        for r in folder_records:

            record = Record(r, folder_key)
            self.records.append(record)


class KeeperFile:

    def __init__(self, f, record_key_bytes):

        self.file_key = ''
        self.meta_dict = None

        self.file_data = None

        self.name = ''
        self.title = ''
        self.type = ''
        self.last_modified = 0
        self.size = 0

        self.f = f
        self.record_key_bytes = record_key_bytes

        # Set file metadata

        meta = self.__get_meta()

        self.title = meta.get('title')
        self.name = meta.get('name')
        self.type = meta.get('type')
        self.last_modified = meta.get('lastModified')
        self.size = meta.get('size')

    def __decrypt_file_key(self):
        file_key_encrypted_base64 = self.f.get('fileKey')
        file_key_encrypted = base64_to_bytes(file_key_encrypted_base64)
        file_key = decrypt_aes(file_key_encrypted, self.record_key_bytes)
        return file_key

    def __get_meta(self):
        """
        Returns file metadata dictionary (file name, title, size, type, etc.)
        """
        if not self.meta_dict:
            file_key = self.__decrypt_file_key()

            meta_json = decrypt_aes(base64_to_bytes(self.f.get('data')), file_key)

            self.meta_dict = json_to_dict(meta_json)

        return self.meta_dict

    def get_file_data(self):
        """
        Return decrypted raw file data
        """
        if not self.file_data:    # cached if nothing
            file_key = self.__decrypt_file_key()
            file_url = self.f.get('url')

            rs = requests.get(file_url)

            file_encrypted_data = rs.content

            self.file_data = decrypt_aes(file_encrypted_data, file_key)

        return self.file_data

    def save_file(self, path, create_folders=False):
        """
        Save decrypted file data to the provided path
        """

        if create_folders:
            os.makedirs(os.path.dirname(path), exist_ok=True)

        file_data = self.get_file_data()

        dir_path = os.path.dirname(os.path.abspath(path))

        if not os.path.exists(dir_path):
            raise KeeperError("No such file or directory %s\nConsider adding `create_folders=True` to `save_file()` "
                              "method " % path)

        file = open(path, "wb")

        file.write(file_data)
        file.close()

        return True

    def __str__(self):
        return "[KeeperFile - name: %s, title: %s]" % (self.name, self.title)



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Secrets Manager
# Copyright 2021 Keeper Security Inc.
# Contact: ops@keepersecurity.com

import base64
import uuid
import json
import time
from collections import deque
from requests import Response as RequestResponse
from keeper_secrets_manager_core.utils import encrypt_aes


class ResponseQueue:

    """Queue up response

    The is a FIFO queue. The queue can be loaded with mock Response instance that
    will be shift off when a request to get_response called.

    """

    def __init__(self, client, post_method=None):

        self.client = client

        # If the post_method is not defined, default to a built in one. All it will do is
        # return the next response in thr queue.
        if post_method is None:
            post_method = self.auto_responder_patch
        self.post_method = post_method

        self.queue = deque()

    def add_response(self, response):
        response.client = self.client
        if self.post_method is not None:
            response.patch_post_query(self.post_method)
        self.queue.append(response)

    def get_response(self, context):
        if len(self.queue) == 0:
            raise ValueError("Not enough queued responses. Cannot get response.")
        response = self.queue.popleft()
        return response.instance(context)

    def auto_responder_patch(self, _path, context, _payload_and_signature):
        return self.get_response(context)


class Response:

    """Mock a response from Secrets Manager Service

    secrets_manager_instance = SecretManger()
    mock_res = MockResponse(client=secrets_manager_instance)
    record = mock_res.add_record(title="My Record")
    record.login = "My Login"
    record.password = "My Password"
    record.add_file(name="My File 1")
    record.add_file(name="My File 2")

    response = mock_res.response()

    """

    def __init__(self, client=None, post_method=None, content=None, status_code=200, reason="OK", flags=None):

        """
        Flags is a dictionary where we can turn on/off stuff in the response to fake qwarks from other application and
        how they might store the data.

        * prune_custom_fields - If there is no custom fields, the 'custom' key is removed from the JSON.

        """

        self.client = client

        self.records = {}
        self.folders = {}

        self.content = content
        self.status_code = status_code
        self.reason = reason
        self.flags = flags

        # Monkey patch the _post_query method with a method that will return a
        # requests Response obj.
        if post_method is not None:
            self.patch_post_query(post_method)

        self.headers = {
            "Server": "keeper",
            "Content-Type": "application/octet-stream",
            "Connection": "keep-alive",
            "X-Frame-Options": "DENY",
            "Strict-Transport-Security": "max-age=31536000; includeSubDomains; preload",
            "X-Content-Type-Options": "nosniff",
            "X-XSS-Protection": "1; mode=block",
            "Expect-CT": "max-age=10, report-uri=\"https://keepersecurity.report-uri.com/r/d/ct/reportOnly\"",
            "Accept-Ranges": "bytes"
        }

    def patch_post_query(self, patch):
        """Patch keeper_secrets_manager.core.SecretsManager._post_query

        The patch is a method that accept three arguments.

          * path - Endpoint Path.
          * context - Context instance.
          * payload_and_signature - Dictionary containing encrypted data and signature.

        When SecretsManager posts a message to the Secrets Manager services, it will call the patch method
        instead. The patch method needs to return an instance of the requests Response.

        """

        if self.client is None:
            raise ValueError("The secrets manager client has not been set.")

        self.client._post_query = patch

    def dump(self, secret, flags=None):

        return {
            "encryptedAppKey": None,
            "folders": [self.folders[uid].dump(secret=secret, flags=flags) for uid in self.folders],
            "records": [self.records[uid].dump(secret=secret, flags=flags) for uid in self.records]
        }

    def instance(self, context):

        """ Return a requests Response instance filled in with mock response message

        res = Response.instance(context)

        The method requires an instance of keeper_secrets_manager_core.dto.payload.Context since that
        information on how to encrypt the response message.

        """

        res = RequestResponse()
        for key in self.headers:
            res.headers[key] = self.headers[key]
        res.headers["Date"] = time.strftime("%a, %d %b %Y %H:%M:%S GMT", time.gmtime())

        # If canned content has not be set, the create content from records/folders.
        if self.content is None:
            json_str = json.dumps(self.dump(secret=context.clientKey, flags=self.flags))
            content = encrypt_aes(json_str.encode(), context.transmissionKey.key)
            res._content = content
            res.headers["Content-Length"] = str(len(content))

            res.status_code = 200
            res.reason = "OK"
        # Else return the canned content. This is useful to mock errors that return plain or json text.
        else:
            content = self.content
            if type(content) is str:
                content = content.encode()
            res._content = content
            res.status_code = self.status_code
            res.reason = self.reason

        return res

    def add_record(self, title=None, record_type=None, uid=None, record=None, keeper_record=None):

        if keeper_record is not None:
            record = Record.convert_keeper_record(keeper_record)
        elif record is None:
            record = Record(title=title, record_type=record_type, uid=uid)

        if isinstance(object, Record.__class__) is False:
            raise ValueError("Record being added to the response is not a "
                             "keeper_secrets_manager_core.mock.Record instance.")

        self.records[record.uid] = record
        return record

    def add_folder(self, uid=None, folder=None):

        if folder is None:
            folder = Folder(uid=uid)

        if isinstance(object, Folder.__class__) is False:
            raise ValueError("Folder being added to the response is not a "
                             "keeper_secrets_manager_core.mock.Folder instance.")

        self.folders[folder.uid] = folder
        return folder


class Folder:

    def __init__(self, uid=None):
        if uid is None:
            uid = str(uuid.uuid4())
        self.uid = uid
        self.records = {}

    def add_record(self, title=None, record_type=None, uid=None, record=None):

        if record is None:
            record = Record(record_type=record_type, uid=uid, title=title)

        if isinstance(object, Record.__class__) is False:
            raise ValueError("Record being added to the response is not a "
                             "keeper_secrets_manager_core.mock.Record instance.")

        self.records[record.uid] = record
        return record

    def dump(self, secret, flags=None):

        return {
            "folderUid": self.uid,
            "folderKey": base64.b64encode(encrypt_aes(secret, secret)).decode(),
            "records": [self.records[uid].dump(secret=secret, flags=flags) for uid in self.records]
        }


class File:

    def __init__(self, name, title=None, content_type=None, url=None, content=None, last_modified=None):
        self.uid = str(uuid.uuid4())
        self.secret_used = None

        self.name = name
        if title is None:
            title = self.name
        self.title = title
        if content_type is None:
            content_type = "text/plain"
        self.content_type = content_type
        if url is None:
            url = "http://localhost/{}".format(self.uid)
        self.url = url
        if content is None:
            content = "ABC123"
        self.content = content
        self.size = len(self.content)
        if last_modified is None:
            last_modified = int(time.time())
        self.last_modified = last_modified

    def downloadable_content(self):

        # The dump method will generate the content that the secrets manager would return. The
        # problem is we won't know the secret here. So the dump method needs to be run before
        # this method is called. The dump method will set the last/only secret used. We need to
        # encode the content with that secret.
        if self.secret_used is None:
            raise ValueError("The file has not be dump'd yet, Secret is unknown.")

        data = self.content
        if type(data) is not bytes:
            data = data.encode()

        return encrypt_aes(data, self.secret_used)

    def dump(self, secret, flags=None):

        # No special flags for download. Do this to make PEP8 happy for unused vars.
        if flags is not None:
            pass

        self.secret_used = secret
        d = {
            "name": self.name,
            "title": self.title,
            "size": self.size,
            "lastModified": self.last_modified,
            "type": self.content_type
        }
        data = json.dumps(d)
        file_data = {
            "fileUid": self.uid,
            "fileKey": base64.b64encode(encrypt_aes(secret, secret)).decode(),
            "data": base64.b64encode(encrypt_aes(data.encode(), secret)).decode(),
            "url": self.url,
            "thumbnailUrl": None
        }
        return file_data


class Record:

    def __init__(self, record_type=None, uid=None, title=None):

        if uid is None:
            uid = str(uuid.uuid4())
        if record_type is None:
            record_type = "login"

        self.uid = uid
        self.record_type = record_type
        self.title = title
        self.is_editable = False
        self.files = {}

        # Some default data
        self._fields = {
            "login": "Login {}".format(uid),
            "password": "******** {}".format(uid),
            "url": "http://localhost/{}".format(uid)
        }
        self._custom_fields = {}

    @staticmethod
    def convert_keeper_record(keeper_record):

        new_record = Record(
            record_type=keeper_record.type,
            uid=keeper_record.uid,
            title=keeper_record.title
        )
        for item in keeper_record.dict.get("fields"):
            new_record.field(
                field_type=item["type"],
                value=item["value"]
            )
        for item in keeper_record.dict.get("custom", []):
            new_record.custom_field(
                label=item["label"],
                field_type=item["type"],
                value=item["value"]
            )
        # TODO - Add files
        return new_record

    def field(self, field_type, value):
        if type(value) is not list:
            value = [value]
        self._fields[field_type] = value

    def custom_field(self, label, value, field_type="text"):
        if type(value) is not list:
            value = [value]
        if label not in self._custom_fields:
            self._custom_fields[label] = {}
        self._custom_fields[label][field_type] = value

    def add_file(self, name, title=None, content_type=None, url=None, content=None, last_modified=None):

        file = File(
            name=name,
            title=title,
            content_type=content_type,
            url=url,
            content=content,
            last_modified=last_modified
        )
        self.files[file.uid] = file
        return file

    def dump(self, secret, flags=None):

        # If no files, the JSON has null
        files = None
        if len(self.files) > 0:
            files = [self.files[uid].dump(secret=secret, flags=flags) for uid in self.files]

        fields = [{"type": "fileRef", "value": [uid for uid in self.files]}]
        for field_type in self._fields:
            fields.append({
                "type": field_type,
                "value": self._fields[field_type]
            })

        custom = []
        for label in self._custom_fields:
            for field_type in self._custom_fields[label]:
                custom.append({
                    "type": field_type,
                    "label": label,
                    "value": self._custom_fields[label][field_type]
                })

        record_data = {
            "title": self.title,
            "type": self.record_type,
            "fields": fields,
            "custom": custom
        }

        if flags is not None:
            # SecretsManager will not add a custom key if there is no custom fields. However the UI does.
            if flags.get("prune_custom_fields", False) is True and len(record_data["custom"]) == 0:
                record_data.pop("custom", None)
            # Remove fields if they do not have values.
            if flags.get("prune_empty_fields", False) is True:
                new_fields = []
                for field in record_data["fields"]:
                    if len(field.get("value")) > 0:
                        new_fields.append(field)
                record_data["fields"] = new_fields

        data = {
            "recordUid": self.uid,
            "recordKey": base64.b64encode(encrypt_aes(secret, secret)).decode(),
            "data": base64.b64encode(encrypt_aes(json.dumps(record_data).encode(), secret)).decode(),
            "isEditable": self.is_editable,
            "files": files
        }

        return data



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Secrets Manager
# Copyright 2021 Keeper Security Inc.
# Contact: ops@keepersecurity.com

class Context:

    def __init__(self, transmission_key, client_id, client_key):
        self.transmissionKey = transmission_key
        self.clientId = client_id
        self.clientKey = client_key


class TransmissionKey:

    def __init__(self, publicKeyId, key, encryptedKey):
        self.publicKeyId = publicKeyId
        self.key = key
        self.encryptedKey = encryptedKey


class GetPayload:

    def __init__(self):
        self.clientVersion = None
        self.clientId = None
        self.publicKey = None
        self.requestedRecords = None


class UpdatePayload:

    def __init__(self):
        self.clientVersion = None
        self.clientId = None
        self.recordUid = None
        self.data = None



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Secrets Manager
# Copyright 2021 Keeper Security Inc.
# Contact: ops@keepersecurity.com

import base64
import os

from Cryptodome.Cipher import AES, PKCS1_v1_5
from Cryptodome.PublicKey import RSA
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ec


class CryptoUtils:

    @staticmethod
    def bytes_to_int(b):
        return int.from_bytes(b, byteorder='big')

    @staticmethod
    def url_safe_str_to_bytes(s):
        b = base64.urlsafe_b64decode(s + '==')
        return b

    @staticmethod
    def generate_random_bytes(length):
        return os.urandom(length)

    @staticmethod
    def generate_encryption_key_bytes():
        """Generates 32 bit transmission key"""
        return CryptoUtils.generate_random_bytes(32)

    @staticmethod
    def bytes_to_url_safe_str(b):
        return base64.urlsafe_b64encode(b).decode().rstrip('=')

    @staticmethod
    def url_safe_str_to_int(s):
        b = CryptoUtils.url_safe_str_to_bytes(s)
        return CryptoUtils.bytes_to_int(b)

    @staticmethod
    def generate_ecc_keys():

        encryption_key_bytes = CryptoUtils.generate_encryption_key_bytes()
        private_key_str = CryptoUtils.bytes_to_url_safe_str(encryption_key_bytes)
        encryption_key_int = CryptoUtils.url_safe_str_to_int(private_key_str)
        private_key = ec.derive_private_key(encryption_key_int, ec.SECP256R1(), default_backend())

        return private_key

    @staticmethod
    def public_key_ecc(private_key):
        pub_key = private_key.public_key()
        pub_key_bytes = pub_key.public_bytes(serialization.Encoding.X962, serialization.PublicFormat.UncompressedPoint)
        return pub_key_bytes

    @staticmethod
    def encrypt_aes(data, key):
        # type: (bytes, bytes) -> bytes
        iv = os.urandom(12)
        cipher = AES.new(key=key, mode=AES.MODE_GCM, nonce=iv)
        enc_data, tag = cipher.encrypt_and_digest(data)
        return iv + enc_data + tag

    @staticmethod
    def encrypt_rsa(data, rsa_key):
        # type: (bytes, RSA.RsaKey) -> bytes
        cipher = PKCS1_v1_5.new(rsa_key)
        return cipher.encrypt(data)


# if __name__ == '__main__':
#
    # pk = CryptoUtils.generate_encryption_key_bytes()
    # print(pk)
